# app/models/concerns/file_relocation_concern.rb
module S3FileHandleable
  extend ActiveSupport::Concern

  def relocate_files_on_folder_name_change
    return true unless saved_change_to_s3_folder_name?
    name_changes = saved_changes[:s3_folder_name]
    return true if name_changes[0].nil?

    Hyrax::HandleS3FolderNameJob.perform_later(self.class.name, self.id, name_changes)
  end

  def handle_meta_file_on_s3
    s3 = S3StorageService.new
    s3.init_client

    crc_dataset = self.class.name == "CrcDataset" ? self.class.find(self.id) : self.crc_dataset
    bucket_name = s3.sanitise_name(crc_dataset.id)
    complex_modalities = self.class.name == "ComplexModality" ? [self] : self.complex_modalities.to_a.flatten

    complex_modalities.each do |modality|
      subject = modality.complex_subject
      session = modality.complex_session
      meta_file_key = "#{subject.s3_folder_name}/#{session.s3_folder_name}/#{modality.s3_folder_name}/meta.json"
      json_data = prepare_json(crc_dataset, subject, session, modality)

      s3.add_content(bucket_name, meta_file_key, json_data)
    end
  end

  private

  def prepare_json(crc_dataset, subject, session, modality)
    @meta = {}

    data_from_crc_dataset(crc_dataset)
    data_from_complex_subject(subject)
    data_from_complex_session(session)
    data_from_complex_modality(modality)

    @meta.to_json
  end

  def data_from_complex_subject(subject)
    @meta["Subject ID"] = subject.id
    @meta["Subject Species"] = subject.subject_species
    @meta["Subject type"] = subject.subject_type
    @meta["Subject sex"] = subject.subject_sex
    @meta["Subject age"] = subject.subject_age
  end

  def data_from_complex_session(session)
    # TODO Date recorded curruntly fetch from ComplexDate we need to confirm is this correct?
  end

  def data_from_complex_modality(modality)
    @meta["Modality"] =  modality.modality
  end

  def data_from_crc_dataset(crc_dataset)
    @meta["DublinCore-Title"] = crc_dataset.title[0]
    @meta["Experiment title"] = crc_dataset.title[0]
    @meta["DataCite-Title"] = crc_dataset.title[0]
    @meta["DublinCore-Language"] = crc_dataset.language[0]
    @meta["DataCite-Language"] = crc_dataset.language[0]
    @meta["Resource Type"] = crc_dataset.resource_type[0]
    @meta["Experiment Description"] = crc_dataset.experiment_description[0]
    @meta["Group ID"] = "" # Complex Identifier removed
    @meta["Animal|Ethics approval No."] = crc_dataset.approval_number
    @meta["DublinCore-Subject"] = crc_dataset.subject.join("; ")
    @meta["DataCite-Subject"] = crc_dataset.subject.join("; ")
    @meta["DublinCore-Publisher"] =  crc_dataset.publisher[0]
    @meta["DataCite-Publisher"] =  crc_dataset.publisher[0]
    @meta["DublinCore-Coverage"] = crc_dataset.coverage
    @meta["Software version"] = "" # TODO we don't have for now
    @meta["Extra information"] = crc_dataset.extra_information[0]

    data_from_complex_person(crc_dataset.complex_person)
    data_from_complex_date(crc_dataset.complex_date)
    data_from_complex_funding_reference(crc_dataset.complex_funding_reference)
  end

  def data_from_complex_person(complex_person)
    return unless complex_person.any?

    complex_person = complex_person.group_by{ |p| p.role[0] }

    unless complex_person["creator"].nil?
      @meta["Creator"] = complex_person["creator"].map{ |creator| "#{creator.first_name[0]}, #{creator.last_name[0]}" }.compact.join("; ")
      @meta["DataCite-creatorName"] = @meta["Creator"]
      @meta["DublinCore-Creator"] = @meta["Creator"]
      @meta["DataCite-creatorFamilyName"] = complex_person["creator"].map{ |creator| creator.first_name[0] }.compact.join("; ")
      @meta["DataCite-creatorGivenName"] = complex_person["creator"].map{ |creator| creator.last_name[0] }.compact.join("; ")
      @meta["DataCite-creatorAffiliation"] = complex_person["creator"].map{ |creator| creator.affiliation[0] }.compact.join("; ")
      @meta["DataCite-creatorNameIdentifier"] = complex_person["creator"].map{ |creator| creator.orcid[0] }.compact.join("; ")
    end

    unless complex_person["contributor"].nil?
      @meta["contributor"] = complex_person["contributor"].map{ |contributor| "#{contributor.first_name[0]}, #{contributor.last_name[0]}" }.compact.join("; ")
      @meta["DataCite-contributorName"] = @meta["contributor"]
      @meta["DublinCore-Contributor"] = @meta["contributor"]
      @meta["DataCite-contributorFamilyName"] = complex_person["contributor"].map{ |contributor| contributor.first_name[0] }.compact.join("; ")
      @meta["DataCite-contributorGivenName"] = complex_person["contributor"].map{ |contributor| contributor.last_name[0] }.compact.join("; ")
      @meta["DataCite-contirbutorAffiliation"] = complex_person["contributor"].map{ |contributor| contributor.affiliation[0] }.compact.join("; ")
      @meta["DataCite-contributorNameIdentifier"] = complex_person["contributor"].map{ |contributor| contributor.orcid[0] }.compact.join("; ")
    end
  end

  def data_from_complex_funding_reference(complex_funding_reference)
    return unless complex_funding_reference.any?

    @meta["DataCite-FundingReference"] = complex_funding_reference.map{ |funding| funding.funder_name[0] }.compact.join("; ")
    @meta["DataCite-funderIdentifier"] = complex_funding_reference.map{ |funding| funding.funder_identifier[0] }.compact.join("; ")
  end

  def data_from_complex_date(complex_dates)
    return unless complex_dates.any?

    @meta["Record date"] = complex_dates.map{ |complex_date| complex_date.date[0] if complex_date.description[0] == "Recorded" }.compact.join("; ")
    @meta["DataCite-dateType"] = complex_dates.map{ |complex_date| complex_date.description[0] if complex_date.description[0] != "Recorded" }.compact.join("; ")
    @meta["DataCite-Date"] = complex_dates.map{ |complex_date| complex_date.date[0] if complex_date.description[0] != "Recorded" }.compact.join("; ")
    @meta["DublinCore-Date"] = @meta["DataCite-Date"]
  end

end
